from dataclasses import dataclass

from . import bkmk_json, xbel, chrome_json, netscape_html
from .base import *

FORMAT_EXTS = {
    ".xbel": ["xbel"],
    ".html": ["netscape-html"],
    ".bkmk.json": ["bkmk-json"],
    ".json": ["chrome-json", "bkmk-json"],
}

FORMATS = {
    "bkmk-json": "bkmk JSON - own custom format, easiest for scripting with jq(1)",
    "xbel": "XML Bookmark Exchange Language 1.0",
    "netscape-html": "NETSCAPE Bookmark file 1 - supported by most browsers including Firefox and Chrome",
    "chrome-json": "Chrome Bookmarks JSON - used internally by Chrome",
}

def guess_format(name, verbose=False):
    fmt = None
    for ext, fmts in FORMAT_EXTS.items():
        if name.endswith(ext):
            fmt = fmts[0]
            if len(fmts) > 1 and verbose:
                log("note: guessing format %s for *%s; possible others are: %s" % (fmt, ext, ", ".join(fmts[1:])))
            break
    if fmt is not None:
        return fmt
    else:
        raise ValueError("could not guess format of %s" % name)

def _to_numid(i):
    if not i: return (0, i)
    try:
        return (int(i), i)
    except ValueError:
        return (10**len(i), i)

def _find_max_id(node):
    iid = _to_numid(node.id)
    if isinstance(node, Folder):
        return max([iid] + [_find_max_id(c) for c in node.children])
    else:
        return iid

def _fill_ids(ref, node):
    if node.id == "":
        node.id = str(ref[0])
        ref[0] += 1
    if isinstance(node, Folder):
        for c in node.children:
            _fill_ids(ref, c)

def _find_special_folders(rem, node):
    if isinstance(node, Folder):
        if node.special:
            rem.pop(node.special, None)
        for c in node.children:
            _find_special_folders(rem, c)

def _fill_timestamps(node, ts):
    if isinstance(node, Separator):
        if node.date_added is None:
            node.date_added = ts
    elif isinstance(node, Bookmark):
        if node.date_added is None:
            node.date_added = ts
        if node.date_modified is None:
            node.date_modified = ts
        if node.url_date_modified is None:
            node.url_date_modified = ts
        if node.url_date_visited is None:
            node.url_date_visited = ts
    elif isinstance(node, Folder):
        if node.date_added is None:
            node.date_added = ts
        if node.date_modified is None:
            node.date_modified = ts
        for c in node.children:
            _fill_timestamps(c, ts)
    else:
        assert False

@dataclass
class Bookmarks:
    root: Folder

    def _debug_eq(self, other, accept, xchildren):
        return self.root._debug_eq(other.root, accept, xchildren)

    def fill_special(self):
        """Fill in missing special folders, so every special folder exists"""
        rem = dict(FOLDER_DEFAULT_NAMES)
        _find_special_folders(rem, self.root)
        root_name = rem.pop(None)
        if not self.root.name:
            self.root.name = root_name
        extra_children = []
        for (special, name) in rem.items():
            extra_children.append(Folder("", None, name, "", None, [], special))
        self.root.children = extra_children + self.root.children

    def fill_ids(self):
        """Fill in missing ids, so every element has an id"""
        max_id = _find_max_id(self.root)
        _fill_ids([max_id[0] + 1], self.root)

    def fill_timestamps(self, ts=None):
        """Fill in missing timestamps, so every element has all timestamps"""
        if ts is None:
            import time
            ts = int(time.time() * 1000000)
        _fill_timestamps(self.root, ts)

    @classmethod
    def new(cls):
        return cls(Folder.new())

    @classmethod
    def read(cls, fp_in, fmt_in, fill_special=False, fill_ids=False, fill_timestamps=False):
        if fmt_in not in FORMATS:
            raise ValueError("not a valid format: %s" % fmt_in)
        bm = cls(globals()[fmt_in.replace("-", "_")].read(fp_in))
        if fill_special:
            bm.fill_special()
        if fill_ids:
            bm.fill_ids()
        if fill_timestamps:
            bm.fill_timestamps()
        return bm

    def write(self, fp_out, fmt_out, cull_special=False, cull_attr=False):
        if fmt_out not in FORMATS:
            raise ValueError("not a valid format: %s" % fmt_out)
        globals()[fmt_out.replace("-", "_")].write(self.root, fp_out, cull_special, cull_attr)

    @staticmethod
    def sanity_check_args(do_warn, **kwargs):
        ok = [True]
        def bad_args(s):
            if do_warn: log("warn:", s)
            ok[0] = False
        if not kwargs["cull_special"] and kwargs["fill_special"]:
            # in a round-trip test, more and more cruft folders will accumulate
            bad_args("cull_special=False may cause fill_special=True to write empty non-special cruft folders")
        if kwargs["cull_attr"] and kwargs["fill_ids"]:
            # in a round-trip test, ids will keep getting deleted and reassigned
            bad_args("cull_attr=True may undo some/all of the effects of fill_ids=True")
        if kwargs["cull_attr"] and kwargs["fill_timestamps"]:
            # in a round-trip test, timestamps will keep getting deleted and reassigned
            bad_args("cull_attr=True may undo some/all of the effects of fill_timestamps=True")
        return ok[0]



from .base import *

import json

def from_ast(node):
    node_type = node["type"]
    id = node.get("id", "")
    date_added = node.get("date_added", None)
    name = node.get("name", "")
    icon = node.get("icon", "")
    date_modified = node.get("date_modified", None)
    if node_type == "folder":
        children = list(map(from_ast, node["children"]))
        special = SpecialFolder[node["special"]] if "special" in node else None
        return Folder(id, date_added, name, icon, date_modified, children, special)
    elif node_type == "bookmark":
        url = node["url"]
        url_date_modified = node.get("url_date_modified", None)
        url_date_visited = node.get("url_date_visited", None)
        return Bookmark(id, date_added, name, icon, date_modified, url, url_date_modified, url_date_visited)
    elif node_type == "separator":
        return Separator(id, date_added)
    else:
        raise ValueError("unrecognised node type: %s" % node_type)

def to_ast(node):
    if isinstance(node, Separator):
        return _d({
            "type": "separator",
            "id": _oe(node.id),
            "date_added": _on(node.date_added),
        })
    elif isinstance(node, Bookmark):
        return _d({
            "type": "bookmark",
            "id": _oe(node.id),
            "date_added": _on(node.date_added),
            "name": node.name,
            "icon": _oe(node.icon),
            "date_modified": _on(node.date_modified),
            "url": node.url,
            "url_date_modified": _on(node.url_date_modified),
            "url_date_visited": _on(node.url_date_visited),
        })
    elif isinstance(node, Folder):
        return _d({
            "type": "folder",
            "id": _oe(node.id),
            "date_added": _on(node.date_added),
            "name": node.name,
            "icon": _oe(node.icon),
            "date_modified": _on(node.date_modified),
            "children": list(map(to_ast, node.children)),
            "special": _o(node.special.name if node.special else None, None),
        })
    else:
        assert False

def read(fp_in):
    return from_ast(json.load(fp_in))

def write(root, fp_out, _cull_attr, _cull_special):
    json.dump(to_ast(root), fp_out, indent=2)

def _roundtrip_acceptable_diff(*args):
    # bkmk-json should support everything, no diff is acceptable
    return False



from .base import *

from datetime import datetime
import xml.etree.ElementTree as ET

"""https://xbel.sourceforge.net/language/versions/1.0/xbel-1.0.xhtml"""
XBEL_VERSION = "1.0"

def from_fmt_time(iso):
    if iso is None: return None
    import sys
    if not (sys.version_info.major >= 3 and sys.version_info.minor >= 11):
        log("warn: parsing ISO 8601 dates properly requires Python >= 3.11")
    return int(datetime.fromisoformat(iso).timestamp() * 1000000)

def to_fmt_time(ue):
    if ue is None: return None
    # utcfromtimestap/isoformat are buggy, we have to add Z ourselves
    iso = datetime.utcfromtimestamp(ue / 1000000.0).isoformat()
    return iso if iso.endswith("Z") else iso + "Z"

# we don't support aliases, nor likely ever will, as it's hard to fit into the other formats
SUPPORTED_TAGS = ["xbel", "folder", "bookmark", "separator"]

SPECIAL_FOLDERS_BY_NAME = {
    "toolbar": SpecialFolder.TOOLBAR,
}

SPECIAL_FOLDERS_BY_ENUM = {v: k for (k, v) in SPECIAL_FOLDERS_BY_NAME.items()}

def from_ast(node):
    node_type = node.tag
    attrib = node.attrib
    id = attrib.get("id", "")
    date_added = from_fmt_time(attrib.get("added", None))
    title = node.find("title")
    name = title.text if title is not None and title.text is not None else ""
    icon = attrib.get("icon", "")
    date_modified = from_fmt_time(None) # TODO: not supported by format
    if node_type == "folder" or node_type == "xbel":
        children = [from_ast(c) for c in node if c.tag in SUPPORTED_TAGS]
        special = SpecialFolder.TOOLBAR if node_type == "folder" and node.get("toolbar", "") == "yes" else None
        return Folder(id, date_added, name, icon, date_modified, children, special)
    elif node_type == "bookmark":
        url = attrib["href"]
        url_date_modified = from_fmt_time(attrib.get("modified", None))
        url_date_visited = from_fmt_time(attrib.get("visited", None))
        if any(url.startswith(f) for f in FAKE_SEPARATOR_URLS):
            # floccus uses fake separators even though xbel supports real ones
            return Separator(id, date_added)
        else:
            return Bookmark(id, date_added, name, icon, date_modified, url, url_date_modified, url_date_visited)
    elif node_type == "separator":
        return Separator(id, date_added)
    else:
        raise ValueError("unrecognised node type: %s" % node_type)

def to_ast(node, cull_special, cull_attr, depth=0):
    if isinstance(node, Separator):
        w = ET.Element("separator")
        w.attrib = _d({
            "id": _oe(node.id),
            "added": _on(to_fmt_time(node.date_added)),
        })
        if cull_attr:
            w.attrib.pop("id", None)
            w.attrib.pop("added", None)
        return w
    elif isinstance(node, Bookmark):
        w = ET.Element("bookmark")
        t = ET.SubElement(w, "title")
        t.text = node.name
        w.attrib = _d({
            "id": _oe(node.id),
            "added": _on(to_fmt_time(node.date_added)),
            "icon": _oe(node.icon),
            # TODO: not supported by format
            #"": _on(to_fmt_time(node.date_modified)),
            "href": node.url,
            "modified": _on(to_fmt_time(node.url_date_modified)),
            "visited": _on(to_fmt_time(node.url_date_visited)),
        })
        return w
    elif isinstance(node, Folder):
        w = ET.Element("xbel" if depth == 0 else "folder")
        t = ET.SubElement(w, "title")
        t.text = node.name
        w.attrib = _d({
            "id": _oe(node.id),
            "added": _on(to_fmt_time(node.date_added)),
            "icon": _oe(node.icon),
            # TODO: not supported by format
            #"": _on(to_fmt_time(node.date_modified)),
            # TODO: other values not supported by format
            "toolbar": _o("yes" if node.special == SpecialFolder.TOOLBAR else None, None),
        })
        if depth == 0:
            w.attrib["version"] = XBEL_VERSION
        if cull_attr and depth == 0:
            w.attrib.pop("icon", None)
        w.extend([
            to_ast(c, cull_special, cull_attr, depth+1) for c in node.children
            if _keep_child(cull_special, SPECIAL_FOLDERS_BY_ENUM.keys(), c)
        ])
        return w
    else:
        assert False

def read(fp_in):
    x = ET.parse(fp_in)
    ver = x.getroot().attrib.get("version", None)
    if ver != XBEL_VERSION:
        raise ValueError("unsupported xbel version: %s" % ver)
    root = from_ast(x.getroot())
    assert isinstance(root, Folder)
    return root

def write(root, fp_out, cull_special, cull_attr):
    w = to_ast(root, cull_special, cull_attr)
    ET.indent(w)
    # we opened the file in string mode, so write in string mode
    ET.ElementTree(w).write(fp_out, encoding="unicode", xml_declaration=True, short_empty_elements=False)

def _roundtrip_acceptable_diff(cull_attr, depth, ty, attr, v_a, v_b):
    if attr == "id":
        if cull_attr and ty == Separator and v_a != "" and v_b == "":
            # separators don't have ids under cull_attr
            return True
    elif attr == "date_added":
        if cull_attr and ty == Separator and v_a is not None and v_b is None:
            # separators don't have date_added under cull_attr
            return True
    elif attr == "date_modified":
        # attribute unsupported by format
        # don't bother checking the value as fill_timestamps sets any value
        return True
    elif attr == "special":
        if v_a is not None and v_a not in SPECIAL_FOLDERS_BY_ENUM.keys() and v_b is None:
            # only some special folders are supported
            return True
    return False



from .base import *

import json

SUPPORTED_VERSION = 1
WINDOWS_UNIX_DIFF = 11644473600000000

def from_fmt_time(cs):
    if cs is None: return None
    return int(cs) - WINDOWS_UNIX_DIFF

def to_fmt_time(ue):
    if ue is None: return None
    return str(ue + WINDOWS_UNIX_DIFF)

# chrome bookmarks have a bit of a customised root structure, deal with it here

SPECIAL_FOLDERS_BY_NAME = {
    "bookmark_bar": SpecialFolder.TOOLBAR,
    "other": SpecialFolder.OTHER_UNFILED,
    "tabs": SpecialFolder.SAVED_TABS,
    "synced": None,
}

SPECIAL_FOLDERS_BY_ENUM = {v: k for (k, v) in SPECIAL_FOLDERS_BY_NAME.items()}

def from_ast(node, special=None):
    node_type = node["type"]
    id = node.get("id", "")
    date_added = from_fmt_time(node.get("date_added", None))
    name = node.get("name", "")
    icon = "" # TODO: not supported by format
    date_modified = from_fmt_time(node.get("date_modified", None))
    if node_type == "folder":
        children = list(map(from_ast, node["children"]))
        return Folder(id, date_added, name, icon, date_modified, children, special)
    elif node_type == "url":
        url = node["url"]
        url_date_modified = None # TODO: not supported by format
        url_date_visited = from_fmt_time(node.get("date_last_used", None))
        if any(url.startswith(f) for f in FAKE_SEPARATOR_URLS):
            return Separator(id, date_added)
        else:
            return Bookmark(id, date_added, name, icon, date_modified, url, url_date_modified, url_date_visited)
    else:
        raise ValueError("unrecognised node type: %s" % node_type)

def to_ast(node, cull_special):
    if isinstance(node, Separator):
        return _d({
            "type": "url",
            "id": _oe(node.id),
            "date_added": _on(to_fmt_time(node.date_added)),
            "name": "|",
            "icon": FAKE_SEPARATOR_ICON,
            "date_modified": _on(to_fmt_time(node.date_added)),
            "url": FAKE_SEPARATOR_URLS[0],
            "date_last_used": _on(to_fmt_time(node.date_added)),
        })
    elif isinstance(node, Bookmark):
        return _d({
            "type": "url",
            "id": _oe(node.id),
            "date_added": _on(to_fmt_time(node.date_added)),
            "name": node.name,
            "icon": _oe(node.icon),
            "date_modified": _on(to_fmt_time(node.date_modified)),
            "url": node.url,
            # TODO: not supported by format
            #"": _on(to_fmt_time(node.url_date_modified)),
            "date_last_used": _on(to_fmt_time(node.url_date_visited)),
        })
    elif isinstance(node, Folder):
        return _d({
            "type": "folder",
            "id": _oe(node.id),
            "date_added": _on(to_fmt_time(node.date_added)),
            "name": node.name,
            "icon": _oe(node.icon),
            "date_modified": _on(to_fmt_time(node.date_modified)),
            "children": [
                to_ast(c, cull_special) for c in node.children
                if _keep_child(cull_special, SPECIAL_FOLDERS_BY_ENUM.keys(), c)
            ],
        })
    else:
        assert False

def read(fp_in):
    p = json.load(fp_in)
    if "version" not in p:
        ver = SUPPORTED_VERSION
        log("warn: no version found, assume we support it")
    else:
        ver = p["version"]
        if ver != SUPPORTED_VERSION:
            raise ValueError("unsupported chrome-json version: %s" % ver)

    w = from_ast(p["roots"][SPECIAL_FOLDERS_BY_ENUM[None]], None)
    assert isinstance(w, Folder)
    special_children = []
    for k, v in p["roots"].items():
        if k == SPECIAL_FOLDERS_BY_ENUM[None]: continue
        root = from_ast(v, SPECIAL_FOLDERS_BY_NAME[k])
        assert isinstance(root, Folder)
        special_children.append(root)
    w.children = special_children + w.children
    return w

def write(root, fp_out, cull_special, _cull_attr):
    w = {
        "version": SUPPORTED_VERSION,
        "roots": {},
    }
    is_special = lambda c: isinstance(c, Folder) and c.special is not None
    special_children = [c for c in root.children if is_special(c)]
    for r in special_children:
        w["roots"][SPECIAL_FOLDERS_BY_ENUM[r.special]] = to_ast(r, cull_special)
    # writing to root is pretty hacky, oh well
    orig_children = root.children
    root.children = [c for c in root.children if not is_special(c)]
    w["roots"][SPECIAL_FOLDERS_BY_ENUM[None]] = to_ast(root, cull_special)
    json.dump(w, fp_out, indent=2)
    root.children = orig_children

def _roundtrip_sortkey(node):
    if isinstance(node, Folder) and node.special:
        return node.special.value
    else:
        return 0

def _roundtrip_acceptable_diff(cull_attr, depth, ty, attr, v_a, v_b):
    if attr == "icon":
        if v_a != "" and v_b == "":
            # attribute unsupported by format
            return True
    elif attr == "url_date_modified":
        # attribute unsupported by format
        # don't bother checking the value as fill_timestamps sets any value
        return True
    elif attr == "special":
        if v_a is not None and v_a not in SPECIAL_FOLDERS_BY_ENUM.keys() and v_b is None:
            # only some special folders are supported
            return True
    return False



from dataclasses import dataclass
from enum import Enum

FAKE_SEPARATOR_URLS = [
    "about:bookmark-separator",
    "https://separator.mayastudios.com/",
    "https://separator.floccus.org/",
]

FAKE_SEPARATOR_ICON = "data:image/svg+xml;charset=utf-8;base64,\
PHN2ZyB2ZXJzaW9uPSIxLjEiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIgdmll\
d0JveD0iMCAwIDE2IDE2Ij48c3R5bGU+bGluZSB7IHN0cm9rZTogI2E3YTA5ZjsgfTwvc3R5bGU+\
PGxpbmUgeDE9IjgiIHgyPSI4IiB5MT0iMCIgeTI9IjE2IiBmaWxsPSJub25lIi8+PC9zdmc+"

def _debug_eq_attr(self, other, accept, depth, attrs):
    r = []
    for attr in attrs.split():
        self_attr = getattr(self, attr)
        other_attr = getattr(other, attr)
        if self_attr != other_attr and not accept(depth, type(self), attr, self_attr, other_attr):
            log("err:", attr, "different:", str(self_attr), "vs", str(other_attr))
            r.append(False)
    return all(r)

@dataclass
class Base:
    id: str
    """When this entry was added."""
    date_added: int | None # unix epoch, micros

    def _debug_eq(self, other, accept, xchildren, depth):
        r = []
        r.append(_debug_eq_attr(self, other, accept, depth, "id date_added"))
        return all(r)

@dataclass
class Separator(Base):
    def _debug_eq(self, other, accept, xchildren, depth):
        if type(self) != type(other):
            log("err: type mismatch:", type(self).__name__, "vs", type(other).__name__)
            return False
        r = []
        r.append(super()._debug_eq(other, accept, xchildren, depth))
        return all(r)

@dataclass
class UserEntry(Base):
    name: str
    icon: str
    """When this entry was modified."""
    date_modified: int | None # unix epoch, micros

    def _debug_eq(self, other, accept, xchildren, depth):
        r = []
        r.append(super()._debug_eq(other, accept, xchildren, depth))
        r.append(_debug_eq_attr(self, other, accept, depth, "name icon date_modified"))
        return all(r)

@dataclass
class Bookmark(UserEntry):
    url: str
    """When the URL was modified."""
    url_date_modified: int | None # unix epoch, micros
    """When the URL was last visited."""
    url_date_visited: int | None # unix epoch, micros

    def _debug_eq(self, other, accept, xchildren, depth):
        if type(self) != type(other):
            log("err: type mismatch:", type(self).__name__, "vs", type(other).__name__)
            return False
        r = []
        r.append(super()._debug_eq(other, accept, xchildren, depth))
        r.append(_debug_eq_attr(self, other, accept, depth, "url url_date_modified url_date_visited"))
        return all(r)

SpecialFolder = Enum("SpecialFolder", "TOOLBAR OTHER_UNFILED SAVED_TABS".split())
FOLDER_DEFAULT_NAMES = {
    None: "Bookmarks",
    SpecialFolder.TOOLBAR: "Bookmarks Toolbar",
    SpecialFolder.OTHER_UNFILED: "Other Bookmarks",
    SpecialFolder.SAVED_TABS: "Tabs collection",
}

def _keep_child(cull_special, supported, node):
    if isinstance(node, Folder):
        # cull empty special folders if they are unsupported
        return not (cull_special and node.special is not None and not node.children and node.special not in supported)
    else:
        return True

@dataclass
class Folder(UserEntry):
    children: list['BookmarksTy']
    special: None | SpecialFolder

    def _debug_eq(self, other, accept, xchildren, depth=0):
        if type(self) != type(other):
            log("err: type mismatch:", type(self).__name__, "vs", type(other).__name__)
            return False
        r = []
        r.append(super()._debug_eq(other, accept, xchildren, depth))
        r.append(_debug_eq_attr(self, other, accept, depth, "special"))
        for a, b in zip(xchildren(self.children, depth), xchildren(other.children, depth)):
            r.append(a._debug_eq(b, accept, xchildren, depth+1))
        return all(r)

    @classmethod
    def new(cls):
        return cls("", None, "", "", None, [], None)

BookmarksTy = Separator | Bookmark | Folder

"""
Utils for having optional keys in dictionary literals

Usage is like

_d({
  "key": _o(value, remove_if_equal_to_this_default)
})
"""
def _d(d):
    return {k: v for (k, v) in d.items() if v != _REMOVE_ME}

def _o(v, d):
    return _REMOVE_ME if v == d else v

def _oe(v):
    return _o(v, "")

def _on(v):
    return _o(v, None)

_REMOVE_ME = object()

def log(*args):
    import sys
    print("bkmk:", *args, file=sys.stderr)

__all__ = [
    "FAKE_SEPARATOR_URLS", "FAKE_SEPARATOR_ICON", "FOLDER_DEFAULT_NAMES",
    "Separator", "Bookmark", "SpecialFolder", "Folder", "_keep_child",
    "_d", "_o", "_oe", "_on",
    "log",
]



from .base import *

import sys

from html import escape
from html.parser import HTMLParser

DOCTYPE = "NETSCAPE-Bookmark-file-1"

def from_fmt_time(u):
    if u is None: return None
    return int(u) * 1000000

def to_fmt_time(ue):
    if ue is None: return None
    return str(int(ue / 1000000.0))

SPECIAL_FOLDERS_BY_NAME = {
    "personal_toolbar_folder": SpecialFolder.TOOLBAR,
    "unfiled_bookmarks_folder": SpecialFolder.OTHER_UNFILED,
}

SPECIAL_FOLDERS_BY_ENUM = {v: k for (k, v) in SPECIAL_FOLDERS_BY_NAME.items()}

def from_special_folder(attrs):
    for a, b in SPECIAL_FOLDERS_BY_NAME.items():
        if attrs.get(a, "") == "true":
            return b
    return None

def to_special_folder(folder):
    attr = SPECIAL_FOLDERS_BY_ENUM.get(folder.special, None)
    return {attr: "true"} if attr is not None else {}

# we write our own custom parser because these files don't have </dt> tags
# which really confuses beautifulsoup and makes it generate incorrect stuff
# like <dt><dt><dt></dt></dt></dt>
class NetscapeHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__(convert_charrefs=True)
        self.started = False
        self.stack = []
        self.result = None

    def current(self):
        return self.stack[-1] if self.stack else None

    def append_child(self, elem):
        cur = self.current()
        assert isinstance(cur, Folder)
        cur.children.append(elem)
        self.stack.append(elem)

    def pop_any_child(self):
        cur = self.current()
        if not isinstance(cur, Folder):
            self.stack.pop()

    def handle_starttag(self, tag, attrs):
        if not self.started:
            raise ValueError("did not see expected DOCTYPE")
        attrs = dict(attrs)
        id = attrs.get("id", "")
        date_added = from_fmt_time(attrs.get("add_date", None))

        if tag in ("h1", "h3"):
            icon = attrs.get("icon", "")
            date_modified = from_fmt_time(attrs.get("last_modified", None))
            special = from_special_folder(attrs) if tag == "h3" else None
            folder = Folder(id, date_added, "", icon, date_modified, [], special)
            if tag == "h1":
                self.stack.append(folder)
            else:
                self.append_child(folder)
        elif tag == "a":
            icon = attrs.get("icon", "")
            date_modified = from_fmt_time(attrs.get("last_modified", None))
            url = attrs["href"]
            if any(url.startswith(f) for f in FAKE_SEPARATOR_URLS):
                # floccus uses fake separators even though netscape-html supports real ones
                b = Separator(id, date_added)
            else:
                url_date_modified = None # TODO: not supported by format
                url_date_visited = from_fmt_time(attrs.get("last_visit", None))
                b = Bookmark(id, date_added, "", icon, date_modified, url, url_date_modified, url_date_visited)
            self.append_child(b)
        elif tag == "hr":
            self.pop_any_child()
            self.append_child(Separator(id, date_added))
        elif tag == "dt":
            # there are no closing <dt> tags. this effectively closes off the previous Bookmark
            # it is a no-op if the current item is a Folder
            self.pop_any_child()
        elif tag == "dl":
            if not isinstance(self.current(), Folder):
                # hack around floccus and possibly other tools not writing <h1>
                assert not self.stack
                self.stack.append(Folder.new())

    def handle_endtag(self, tag):
        if tag == "dl":
            if not isinstance(self.current(), Folder):
                self.stack.pop()
            assert isinstance(self.current(), Folder)
            n = self.stack.pop()
            if not self.stack:
                assert not self.result
                self.result = n

    def handle_comment(self, data):
        pass

    def handle_decl(self, data):
        if data == "DOCTYPE " + DOCTYPE:
            self.started = True

    def handle_data(self, data):
        cur = self.current()
        if not cur: return
        data = data.strip()
        if not data: return
        if isinstance(cur, Separator): return # floccus and other fake separators
        # sometimes the data is split into multiple chunks
        cur.name += data

def read(fp_in):
    parser = NetscapeHTMLParser()
    parser.feed(fp_in.read())
    result = parser.result
    if result is None:
        raise ValueError("failed to parse anything out of the file")
    return result

def expand_attrs(attrs):
    return "".join(' %s="%s"' % (k.upper(), escape(v, quote=True)) for k, v in attrs.items())

def write_node(node, fp_out, cull_special, cull_attr, depth=0):
    indent = "    " * depth
    if isinstance(node, Separator):
        attrs = _d({
            "id": _oe(node.id),
            "add_date": _on(to_fmt_time(node.date_added)),
        })
        if cull_attr:
            attrs.pop("id", None)
            attrs.pop("add_date", None)
        print("%s<HR%s>" % (indent, expand_attrs(attrs)), file=fp_out)
    elif isinstance(node, Bookmark):
        attrs = _d({
            "id": _oe(node.id),
            "add_date": _on(to_fmt_time(node.date_added)),
            "icon": _oe(node.icon),
            "last_modified": _on(to_fmt_time(node.date_modified)),
            "href": node.url,
            # TODO: not supported by format
            #"": _on(to_fmt_time(node.url_date_modified)),
            "last_visit": _on(to_fmt_time(node.url_date_visited)),
        })
        if cull_attr:
            attrs.pop("id", None)
        print("%s<DT><A%s>%s</A>" % (indent, expand_attrs(attrs), escape(node.name)), file=fp_out)
    elif isinstance(node, Folder):
        attrs = _d({
            "id": _oe(node.id),
            "add_date": _on(to_fmt_time(node.date_added)),
            "icon": _oe(node.icon),
            "last_modified": _on(to_fmt_time(node.date_modified)),
            **to_special_folder(node),
        })
        if cull_attr:
            attrs.pop("id", None)
            attrs.pop("icon", None)
        if depth == 0:
            if cull_attr:
                attrs = {}
            print("<TITLE>%s</TITLE>" % escape(node.name), file=fp_out)
            print("<H1%s>%s</H1>" % (expand_attrs(attrs), escape(node.name)), file=fp_out)
        else:
            print("%s<DT><H3%s>%s</H3>" % (indent, expand_attrs(attrs), escape(node.name)), file=fp_out)
        print("%s<DL><p>" % indent, file=fp_out)
        for c in node.children:
            if _keep_child(cull_special, SPECIAL_FOLDERS_BY_ENUM.keys(), c):
                write_node(c, fp_out, cull_special, cull_attr, depth+1)
        print("%s</DL><p>" % indent, file=fp_out)
    else:
        assert False

def write(root, fp_out, cull_special, cull_attr):
    print("<!DOCTYPE %s>" % DOCTYPE, file=fp_out)
    print('<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=UTF-8">', file=fp_out)
    write_node(root, fp_out, cull_special, cull_attr, 0)

def _roundtrip_acceptable_diff(cull_attr, depth, ty, attr, v_a, v_b):
    if attr == "id":
        if cull_attr and v_a != "" and v_b == "":
            # nothing has ids under cull_atr
            return True
    elif attr == "date_added":
        if cull_attr and ty == Separator and v_a is not None and v_b is None:
            # separators don't have date_added under cull_attr
            return True
    elif attr == "url_date_modified":
        # attribute unsupported by format
        # don't bother checking the value as fill_timestamps sets any value
        return True
    elif attr == "special":
        if v_a is not None and v_a not in SPECIAL_FOLDERS_BY_ENUM.keys() and v_b is None:
            # only some special folders are supported
            return True
    # special netscape quirks
    if attr.startswith("date_") or attr.startswith("url_date_"):
        if v_a is not None and v_b is not None:
            # lack of precision in format
            return to_fmt_time(v_a) == to_fmt_time(v_b)
        if cull_attr and depth == 0 and ty == Folder and v_a is not None and v_b is None:
            # top-level folder can't have any attributes under cull_attr
            return True
    return False



#!/usr/bin/python3

from bkmk import *

import argparse
import contextlib

import os.path
import sys


def file_or_std(stack, path, fmt, flag, stdname, std):
    if path and path != '-':
        fp = stack.enter_context(open(path))
        fmt = fmt if fmt else guess_format(path, verbose=True)
    else:
        fp = std
        if fmt is None:
            raise ValueError("must give explicit %s when using %s" % (flag, stdname))
    return (fp, fmt)

def _real_main(_prog, *argv):
    parser = argparse.ArgumentParser(
        prog = "bkmk",
        description = 'Convert between different bookmark formats')

    parser.add_argument(
        'input', nargs='?', default=None, help="Input path; omit or '-' for stdin.")
    parser.add_argument(
        'output', nargs='?', default=None, help="Output path; omit or '-' for stdout.")
    parser.add_argument(
        '-f', '--from', metavar='FMT', dest='fmt_in', default=None, choices=FORMATS.keys(),
        help="Input format, one of: %s. Omit to auto-detect from input path." % ", ".join(FORMATS.keys()))
    parser.add_argument(
        '-t', '--to', metavar='FMT', dest='fmt_out', default=None, choices=FORMATS.keys(),
        help="Output format, one of: %s. Omit to auto-detect from output path." % ", ".join(FORMATS.keys()))
    parser.add_argument(
        '--fill-special', default=False, action=argparse.BooleanOptionalAction,
        help="Fill in missing special folders after reading, so every special folder exists")
    parser.add_argument(
        '--fill-ids', default=False, action=argparse.BooleanOptionalAction,
        help="Fill in missing ids after reading, so every element has an id")
    parser.add_argument(
        '--fill-timestamps', default=False, action=argparse.BooleanOptionalAction,
        help="Fill in missing timestamps after reading, so every element has all timestamps")
    parser.add_argument(
        '--cull-special', default=False, action=argparse.BooleanOptionalAction,
        help="Cull empty special folders that are not recognised by the output format")
    parser.add_argument(
        '--cull-attr', default=False, action=argparse.BooleanOptionalAction,
        help="""Cull attributes that are not recognised in-context by the output format. \
        That is, attributes that are supported by the format but are attached to an \
        inappropriate element. We never emit attributes if the format does not support \
        it for any element, since that requires making up an attribute name from thin air.
        """)
    args = parser.parse_args(argv)

    with contextlib.ExitStack() as stack:
        fp_in, fmt_in = file_or_std(stack, args.input, args.fmt_in, "-f", "stdin", sys.stdin)
        fp_out, fmt_out = file_or_std(stack, args.output, args.fmt_out, "-t", "stdout", sys.stdout)
        Bookmarks.sanity_check_args(True, **args.__dict__)
        Bookmarks.read(
            fp_in, fmt_in,
            **{k: getattr(args, k) for k in ("fill_special", "fill_ids", "fill_timestamps")}
        ).write(
            fp_out, fmt_out,
            **{k: getattr(args, k) for k in ("cull_special", "cull_attr")}
        )
    return 0

def main():
    return _real_main(*sys.argv)

if __name__ == "__main__":
    sys.exit(main())
